/*
   Siag, Scheme In A Grid
   Copyright (C) 1996  Ulric Eriksson <ulric@edu.stockholm.se>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

/*
 * matrix.c
 *
 * This module hides the details of stringpool and matrix handling.
 */

#include <stdio.h>
#include <string.h>
#include <time.h>
#include "core.h"


/* make sure the cell is allocated */
static void alloc_cell(buffer_page *bp, int row, int col)
{
  int i;

  if (0 == bp->nent_rows) {
    bp->rows = g_new(row_desc, row+1);
    g_assert(bp->rows);
    bp->nent_rows = row+1;
    for (i=0; i<row; i++) {
      bp->rows[i].height = bp->standard_height;
      bp->rows[i].default_features = NULL;
      bp->rows[i].nent_cells = 0;
      bp->rows[i].cells = NULL;
    }
  } else if (row >= bp->nent_rows) {
    bp->rows = g_realloc(bp->rows, (row+1)* sizeof(row_desc));
    g_assert(bp->rows);
    for (i=bp->nent_rows+1; i<=row; i++) {
      bp->rows[i].height = bp->standard_height;
      bp->rows[i].default_features = NULL;
      bp->rows[i].nent_cells = 0;
      bp->rows[i].cells = NULL;
    }
    bp->nent_rows = row+1;
  }

  if (bp->rows[row].nent_cells == 0) {
    bp->rows[row].cells = g_new(cell, col+1);
    g_assert(bp->rows[row].cells);
    for (i=0; i<=row; i++) {
      bp->rows[row].cells[i].formula = NULL;
      bp->rows[row].cells[i].value.number  = 0.0;
      bp->rows[row].cells[i].interpreter = 0;
      bp->rows[row].cells[i].ctype = CELL_TYPE_EMPTY;
      bp->rows[row].cells[i].vtype = VALUE_TYPE_EMPTY;
      bp->rows[row].cells[i].features = NULL;
    }
    bp->rows[row].nent_cells = col+1;
  } else if (col >= bp->rows[row].nent_cells) {
    bp->rows[row].cells = g_realloc(bp->rows[row].cells, 
				    (col+1) * sizeof(cell));
    g_assert(bp->rows[row].cells);
    for (i=bp->rows[row].nent_cells; i<=col; i++) {
      bp->rows[row].cells[i].formula = NULL;
      bp->rows[row].cells[i].value.number  = 0.0;
      bp->rows[row].cells[i].interpreter = 0;
      bp->rows[row].cells[i].ctype = CELL_TYPE_EMPTY;
      bp->rows[row].cells[i].vtype = VALUE_TYPE_EMPTY;
      bp->rows[row].cells[i].features = NULL;
    }
    bp->rows[row].nent_cells = col+1;
  }
}

/* write a cell into the sheet */
/* supposed to return whether the cell was changed, 
   but I think that it's usually wrong as written */
gint ins_data(buffer_page *bp, guchar interpreter, 
	      gchar *formula,  CellType ct, 
	      cval v, ValueType vt,
	      gint row, int col)
{
  int change = FALSE;
  cell *mrowcol;	
  
  if (formula == NULL || strlen(formula) == 0) {
    ct = CELL_TYPE_EMPTY;
    vt = VALUE_TYPE_EMPTY;
  }
  alloc_cell(bp, row, col);                /* allocate cell */
  mrowcol = &(bp->rows[row].cells[col]);   /* find cell */

  if (mrowcol->formula != formula) {	        /* free old formula */
    if (mrowcol->formula) g_free(mrowcol->formula);    /* if needed */
    mrowcol->formula = g_strdup(formula);      /* store new formula */
  }
  
  if (mrowcol->vtype == VALUE_TYPE_STRING)	/* free the old value */
    g_free(mrowcol->value.text);                /* if it was a string */

  if (vt == VALUE_TYPE_STRING) {                /* store new value */
    if (v.text) mrowcol->value.text = g_strdup(v.text);
    else mrowcol->value.text = g_strdup("");
  } else {
    mrowcol->value.number = v.number;
  }
   
  mrowcol->interpreter = interpreter;
  mrowcol->ctype = ct;
  mrowcol->vtype = vt;
  return change;
}				/* ins_data */

gint ins_features(buffer_page *bp, gint row, gint col, CellFeatures *features)
{
/*   if (row > BUFFER_ROWS || col > BUFFER_COLS) return FALSE; */
  
  alloc_cell(bp, row, col);
  
  bp->rows[row].cells[col].features = features;
  return 1;
}

/* check if a cell has been allocated */
static int is_alloc(buffer_page *bp, gint row, gint col)
{
/*   if (row > BUFFER_ROWS || col > BUFFER_COLS) return 0; */
  if (bp->nent_rows <= row) return 0;
  if (bp->rows[row].nent_cells <= col) return 0;
  return 1;
}

gchar *ret_formula(buffer_page *bp, gint row, gint col)
{
  if (!is_alloc(bp, row, col)) return NULL;
  return bp->rows[row].cells[col].formula;
}	/* ret_text */

gint ret_interpreter(buffer_page *bp, gint row, gint col)
{
  if (!is_alloc(bp, row, col)) return 0;
  return bp->rows[row].cells[col].interpreter;
}

gchar *ret_taddr(buffer_page *bp, gint row, gint col)
{
  return ret_formula(bp, row, col);
}

cval ret_val(buffer_page *bp, gint row, gint col)
{
  if (!is_alloc(bp, row, col)) return 0.0;
  return bp->rows[row].cells[col].value;
}				/* ret_val */

gchar *ret_string(buffer_page *bp, gint row, gint col)
{
  cell *mrow;
  char *v;

  if (!is_alloc(bp, row, col)) return NULL;
  mrow = bp->rows[row].cells;
  if (mrow[col].vtype == VALUE_TYPE_STRING) {
    v = mrow[col].value.text;
    return v;
  }
  return NULL;
}

CellType ret_ctype(buffer_page *bp, gint row, gint col)
{
  if (!is_alloc(bp, row, col)) return CELL_TYPE_EMPTY;
  return bp->rows[row].cells[col].ctype;
} 

ValueType ret_vtype(buffer_page *bp, gint row, gint col)
{
  if (!is_alloc(bp, row, col)) return VALUE_TYPE_EMPTY;
  return bp->rows[row].cells[col].vtype;
} 

#ifndef ABS
#define ABS(x) ((x)>0?(x):-(x))
#endif

gchar *ret_pvalue(gchar *tbuf, buffer_page *bp, 
		  gint row, gint col, gint mode)
{
  static char p[1024];
  ValueType vtype;
  double f;

  f = 0.0;

  if (tbuf == NULL) tbuf = p;
/*   if (row > BUFFER_ROWS || col > BUFFER_COLS) return NULL; */
  vtype = ret_vtype(bp, row, col);

  if (vtype == VALUE_TYPE_NUMBER) {
    f = ret_val(bp, row, col).number;
  }
  tbuf[0] = '\0';
  switch (vtype) {
  case VALUE_TYPE_PIXMAP: /* unimplemented */
  case VALUE_TYPE_WIDGET: /* unimplemented */
  case VALUE_TYPE_UNCALCULATED: /* FIXME */
  case VALUE_TYPE_EMPTY:
    break;
  case VALUE_TYPE_ERROR:
    strcpy(tbuf, "ERROR");
    break;
  case VALUE_TYPE_STRING:
    strncpy(tbuf, ret_string(bp, row, col), 1000);
    break;
  case VALUE_TYPE_NUMBER:
    sprintf(tbuf, "%.3f", f);
    break;
  }
  return tbuf;
}

#if 0
gchar *ret_pvalue(gchar *tbuf, buffer_page *bp, 
		  gint row, gint col, gint mode)
{
  char *r;
  r = ret_formula(bp, row, col);
  if (!r) return "";
  if (tbuf) strcpy(tbuf, r);
  return r;
}
#endif
gint row_last_used(buffer_page *bp)
{
  register int r;
  for (r = bp->nent_rows; r >= 1 && bp->rows[r].nent_cells == 0; r--);
  return r;
}

gint col_last_used(buffer_page *bp, gint row)
{
  register int i;
  if (row < 1 || row > bp->nent_rows) return 0;
  for (i = bp->rows[row].nent_cells;
       i >= 1 && ret_ctype(bp, row, i) == CELL_TYPE_EMPTY; i--);
  return i;
}

#if 0
void downshift_matrix(buffer *b, int row)
{
  int i;

  /* free the last line, if there is anything in it */
  if (b->alloc_lines > BUFFER_ROWS && b->matrix[BUFFER_ROWS])
    cfree(b->matrix[BUFFER_ROWS]);

  /* extend the line allocation in the matrix */
  if (b->alloc_lines >= row && b->alloc_lines < BUFFER_ROWS) {
    b->alloc_lines++;
    b->matrix = (spread **)realloc(b->matrix, (b->alloc_lines+1)*sizeof(spread *));
    b->alloc_cols = (long *)realloc(b->alloc_cols, (b->alloc_lines+1)*sizeof(long));
  }

  /* extend the height table */
  if (b->used_lines >= row && b->used_lines < BUFFER_ROWS) {
    b->used_lines++;
    b->height = (int *)realloc(b->height, (b->used_lines+1)*sizeof(int));
  }

  /* move all the stuff down one line */
  for (i = b->alloc_lines; i > row; i--) {
    b->matrix[i] = b->matrix[i - 1];
    b->alloc_cols[i] = b->alloc_cols[i - 1];
  }
  for (i = b->used_lines; i > row; i--)
    b->height[i] = b->height[i - 1];

  /* make the first line point to nothing */
  if (b->alloc_lines >= row) {
    b->matrix[row] = NULL;
    b->alloc_cols[row] = 0;
  }
  if (b->used_lines >= row)
    b->height[row] = 20;

  update_all_references(b, row, 1, BUFFER_ROWS, BUFFER_COLS, 1, 0);

  b->recalc = b->change = TRUE;
}

void upshift_matrix(buffer *b, int row)
{
  int i;

  /* free the first line, if there is anything in it */
  if (b->alloc_lines > row && b->matrix[row])
    cfree(b->matrix[row]);

  /* we don't bother to shrink the matrix or height table */

  /* move everything up one line */
  for (i = row; i < b->alloc_lines; i++) {
    b->matrix[i] = b->matrix[i + 1];
    b->alloc_cols[i] = b->alloc_cols[i + 1];
  }
  for (i = row; i < b->used_lines; i++)
    b->height[i] = b->height[i + 1];

  /* make the last line point to nothing */
  if (b->alloc_lines > row) {
    b->matrix[b->alloc_lines] = NULL;
    b->alloc_cols[b->alloc_lines] = 0;
  }
  if (b->used_lines > row)
    b->height[b->used_lines] = 20;

  update_all_references(b, row, 1, BUFFER_ROWS, BUFFER_COLS, -1, 0);

  b->recalc = b->change = TRUE;
}

void rightshift_matrix(buffer *b, int col)
{
  int i, j;
  /* extend the width table if necessary */
  if (b->used_cols >= col && b->used_cols < BUFFER_COLS) {
    b->used_cols++;
    b->width = (int *)realloc(b->width, (b->used_cols+1)*sizeof(int));
    for (i = b->used_cols; i > col; i--)
      b->width[i] = b->width[i-1];
    b->width[col] = 80;
  }

  /* process all the allocated lines */
  for (i = 1; i <= b->alloc_lines; i++) {
    if (b->matrix[i]) {
      /* check if it needs to grow */
      if (b->alloc_cols[i] >= col) {
	b->alloc_cols[i]++;
	b->matrix[i] = (spread *)realloc(b->matrix[i],
					 (b->alloc_cols[i]+1)*sizeof(spread));
      }

      /* shift to the right */
      for (j = b->alloc_cols[i]; j > col; j--) {
	b->matrix[i][j] = b->matrix[i][j - 1];
      }

      /* and make the leftmost column empty */
      if (b->alloc_cols[i] > col) {
	b->matrix[i][col] = empty_cell;
      }
    }
  }

  update_all_references(b, 1, col, BUFFER_ROWS, BUFFER_COLS, 0, 1);

  b->recalc = b->change = TRUE;
}

void leftshift_matrix(buffer *b, int col)
{
  int i, j;

  /* do not shrink the width table */
  for (j = col; j < b->used_cols; j++)
    b->width[j] = b->width[j + 1];
  if (b->used_cols >= col)
    b->width[col] = 80;

  for (i = 1; i <= b->alloc_lines; i++) {
    if (b->matrix[i]) {
      /* do not free unused columns */

      /* shift to the left */
      for (j = col; j < b->alloc_cols[i]; j++) {
	b->matrix[i][j] = b->matrix[i][j + 1];
      }

      /* and make the rightmost empty */
      if (b->alloc_cols[i] >= col)
	b->matrix[i][b->alloc_cols[i]] = empty_cell;
    }
  }

  update_all_references(b, 1, col, BUFFER_ROWS, BUFFER_COLS, 0, -1);

  b->recalc = b->change = TRUE;
}

#endif
