#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "gui.h"


static gint ng_col_width(GtkNGrid *ng, gpointer data, gint col) 
{
  if (col == 5) return 0;
  return 70;
}
static gint ng_row_height(GtkNGrid *ng, gpointer data, gint row) 
{
  return 22;
}
static gint ng_type(GtkNGrid *ng, gpointer data, gint row, gint col) 
{
  return NGRID_TEXT;
}

static gchar *ng_text(GtkNGrid *ngrid, gpointer data, gint row, gint col) 
{
  notepage *npage;
  buffer_page *bp;
  static char s[1024];
  npage = (notepage *) gtk_object_get_user_data (GTK_OBJECT (ngrid));
  if (0 == row && 0 == col)
    return npage->buf->name;
  if (row == 0) {
    sprintf(s, "%d", col);
    return s;
  }
  if (col == 0) {
    sprintf(s, "%d", row);
    return s;
  }
    
  bp = npage->bpage;
  s[0] = '\0';
  ret_pvalue(s, bp, row, col, -1);
  return s;
}

static GtkNGridEntryFeatures *ng_features(GtkNGrid *ng, gpointer data, 
					  gint row, gint col) 
{
  static GtkNGridEntryFeatures prot_features;
  static GtkNGridEntryFeatures features;

  if (features.gc == NULL) {
    GdkColormap *cmap;
    GdkColor temp_color;
    features.format = GTK_JUSTIFY_CENTER;
    features.highlight = 1;
    prot_features.highlight = 1;
    features.free_me = 0;
    features.font = NULL;
    
    features.gc = gdk_gc_new(GTK_WIDGET(ng)->window);
    prot_features.gc = gdk_gc_new(GTK_WIDGET(ng)->window);

    cmap = gtk_widget_get_colormap(GTK_WIDGET(ng));
    gdk_color_parse("Red", &temp_color);
    gdk_color_alloc(cmap, &temp_color);
    gdk_gc_set_foreground(features.gc, &temp_color);
    gdk_color_parse("Blue", &temp_color);
    gdk_color_alloc(cmap, &temp_color);
    gdk_gc_set_background(features.gc, &temp_color);

    gdk_color_parse("Black", &temp_color);
    gdk_color_alloc(cmap, &temp_color);
    gdk_gc_set_foreground(prot_features.gc, &temp_color);
    gdk_color_parse("Gray", &temp_color);
    gdk_color_alloc(cmap, &temp_color);
    gdk_gc_set_background(prot_features.gc, &temp_color);
  }
  if (row==0 || col==0)
    return &prot_features;
  else if (1 & row & col) 
    return &features;
  else
    return NULL;
}

gint
ng_ngrid_events (GtkWidget *ngrid,
		 GdkEvent  *event)
{
  notepage *npage;
  GdkEventButton *bevent;
  GdkEventKey *kevent;
  gint tx, ty;
  GdkModifierType tmask;
  guint state = 0;

  npage = (notepage *) gtk_object_get_user_data (GTK_OBJECT (ngrid));

  gdk_window_get_pointer (ngrid->window, &tx, &ty, &tmask);
/*   if (!GTK_WIDGET_HAS_FOCUS(ngrid)) return FALSE; */

  switch (event->type)
    {
      gint row, col;
      guint keyval;

    case GDK_KEY_PRESS:
      return TRUE;
      break;
    case GDK_KEY_RELEASE:
      kevent = (GdkEventKey *) event;
      state = kevent->state;
      keyval = kevent->keyval;
      row = GTK_NGRID(ngrid)->point_row;
      col = GTK_NGRID(ngrid)->point_col;
      switch (keyval) 
	{
	  /* FIXME */
	case GDK_Left:
	  npage_set_point(npage, row, MAX(col-1, 1)); 
	  gtk_widget_grab_focus(ngrid);
	  break;
	case GDK_Right:
	  npage_set_point(npage, row, col+1);
	  gtk_widget_grab_focus(ngrid);
	  break;
	case GDK_Up:
	  npage_set_point(npage, MAX(row-1, 1), col);
	  gtk_widget_grab_focus(GTK_WIDGET(npage->ngrid));
	  break;
	case GDK_Down:
	  npage_set_point(npage, row+1, col);
	  gtk_widget_grab_focus(GTK_WIDGET(npage->ngrid));
	  break;
	case GDK_equal:
	  npage_set_point(npage, row, col);
	  npage->state = PAGE_STATE_ENTERING;
	  gtk_widget_grab_focus(GTK_WIDGET(npage->entry));
	  break;
	default:
	  break;
	}
      return TRUE;
      break;
    case GDK_BUTTON_PRESS:
      bevent = (GdkEventButton *) event;
      state = bevent->state;
      gtk_ngrid_coord_to_rowcol(GTK_NGRID(ngrid),
				tx, ty,
			       &row, &col);

      switch (bevent->button)
        {
        case 1:
	  if (row > 0 && col > 0) {
	    npage_set_point(npage, row, col);
	    npage->state = PAGE_STATE_ENTERING;
	    gtk_widget_grab_focus(GTK_WIDGET(npage->entry));
	  }
	}
      return TRUE;
      break;
    default:
      break;
    }
  return FALSE;
}

void assimilate_ngrid_test(GtkNGrid *ngrid) 
{
  GTK_NGRID(ngrid)->col_width = ng_col_width;
  GTK_NGRID(ngrid)->row_height = ng_row_height;
  GTK_NGRID(ngrid)->type = ng_type;
  GTK_NGRID(ngrid)->text = ng_text;
  GTK_NGRID(ngrid)->features = ng_features;
  GTK_NGRID(ngrid)->prot_row = 1;
  GTK_NGRID(ngrid)->prot_col = 1;
  gtk_signal_connect (GTK_OBJECT (ngrid), "event",
                      (GtkSignalFunc) ng_ngrid_events,
                      ngrid);

}


void assimilate_ngrid(GtkNGrid *ngrid) 
{
  GTK_NGRID(ngrid)->col_width = ng_col_width;
  GTK_NGRID(ngrid)->row_height = ng_row_height;
  GTK_NGRID(ngrid)->type = ng_type;
  GTK_NGRID(ngrid)->text = ng_text;
  GTK_NGRID(ngrid)->features = ng_features;
  GTK_NGRID(ngrid)->prot_row = 1;
  GTK_NGRID(ngrid)->prot_col = 1;
  gtk_signal_connect (GTK_OBJECT (ngrid), "event",
                      (GtkSignalFunc) ng_ngrid_events,
                      ngrid);

}
